#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <pthread.h>
#include <errno.h>

#include "osa.h"

osa_mutex_type *osa_mutex_create(void)
{
  osa_mutex_type *m=NULL ;
  pthread_mutexattr_t mutexattr ;

  m=(osa_mutex_type *)calloc(1, sizeof(osa_mutex_type)) ;
  if(m!=NULL)
    {
      if(pthread_cond_init(&(m->cond), NULL))
	{
	  fprintf(stderr,"osa_mutex_create: ERROR: pthread_cond_init failed!\n") ;
	  free(m) ; m=NULL ; exit(-1) ;
	}
      
      pthread_mutexattr_init(&mutexattr) ;
      
#if !defined(__USE_UNIX98) && defined(LINUX)
      /* compiler warning: pthread_mutexattr_setkind_np is not declared in any header. */
      pthread_mutexattr_setkind_np(&mutexattr,PTHREAD_MUTEX_RECURSIVE_NP) ;
#else
      pthread_mutexattr_settype(&mutexattr,PTHREAD_MUTEX_RECURSIVE) ;
#endif
      if(pthread_mutex_init(&(m->mutex), &mutexattr))
	{
	  fprintf(stderr,"osa_mutex_create: ERROR: pthread_mutex_init failed!\n") ;
	  free(m) ; m=NULL ; exit(-1) ;
	}
      
      pthread_mutexattr_destroy(&mutexattr) ;
    }
  else
    {
      exit(-1) ;
    }
        
  return(m) ;
}

int osa_mutex_give(osa_mutex_type *m)
{
  int r=0 ;

  fprintf(stderr,"osa_mutex_give: ENTER m=%p\n",m) ;

  r=pthread_mutex_lock(&(m->mutex)) ;
  if(r==0)
    {
      if(m->owned)
	{
	  m->owned=0 ;
	  do
	    {
	      r=pthread_cond_signal(&(m->cond)) ;
	    }
	  while(r==EINTR) ;
	}
      pthread_mutex_unlock(&(m->mutex)) ;
    }

  fprintf(stderr,"osa_mutex_give: EXIT m=%p, result=%d\n",m,r) ;

  return(r) ;
}


int osa_mutex_take_timed(osa_mutex_type *m, struct timeval *timeout)
{
  int r=0 ;
  struct timespec t ;
  struct timeval now ;

  fprintf(stderr,"osa_mutex_take_timed: ENTER m=%p\n",m) ;

  gettimeofday(&now, NULL) ;
 
  t.tv_sec=now.tv_sec+timeout->tv_sec ;
  t.tv_nsec=(now.tv_usec*1000)+(timeout->tv_usec*1000) ;
  if(t.tv_nsec>(1000*1000*1000))
    {
      t.tv_nsec-=(1000*1000*1000) ;
      t.tv_sec++ ;
    }
  
  r=pthread_mutex_lock(&(m->mutex)) ;
  fprintf(stderr, "osa_mutex_take_timed(): pthread_mutex_lock() returned %d, errno = %d\n", r, errno) ;
  
  if(r==0)
    {
      /* only wait on cond if someone else has the mutex, otherwise, we
       * just took it */
      if(m->owned)
	do
	  {
	    r=pthread_cond_timedwait(&(m->cond), &(m->mutex), &t) ;
	  }
	while(r==EINTR) ;
      
      if(r==ETIMEDOUT)
	r=-3 ;
      else
	{
	  if(r==0)
	    m->owned=1 ;
	}
      pthread_mutex_unlock(&(m->mutex)) ;
    }

  fprintf(stderr,"osa_mutex_take_timed: EXIT m=%p, result=%d\n",m,r) ;

  return(r) ;
}
